/* smvr.c 3/14/1999 Oguz Yetkin (oguz@writeme.com 
   Simple Minded Virtual Reality (SMVR) server
   Cave Version

   WARNING: Keep this file in pure ANSI C for Cave compatibility
	    Do not pollute with C++.
	    (this library was written to separate the libraries
	     from C++ in the first place)

   responds to text requests, loading Wavefront objects
   assigning them names, translating, scaling, and displaying.

   This program initially written for CS774 to separate
   object loading and display from the program logic.

   The following operations are legal within their blocks.
   The blocks must be executed in order.

   1) Initialization State
   
   #>objects <num>  
  
   #<num> is an integer specifying number of maximum 
   #objects for this session 
      
   >init 
  
   >load <objectname> <filename>
   
   load can be repeated <num> times.  
   Loads an object from a wavefront .obj file
   load will overwrite previous definition if it exists.
   >endinit

   2) Display State 
  
    >locate <objectname> <x> <y> <z>
    
    >display <objectname> #defaults to origin 
    >display <objectname> <x> <y> <z> 
    >display <objectname> <x> <y> <z> <sx> <sy> <sz> <ro> <theta> <phi> 
    >display <objectname> <x> <y> <z> <sx> <sy> <sz> <ro> <theta> <phi> <r> <g> <b>
 
   where <objectname> is the string assigned to the object
   in block 2, and <x> <y> <z> are coordinates of the object
   In the cave, x E [-5,5] y E [0,10] z E [-5,5].  Positive z is
   towards the user, x is horizontal, and y is vertical

   >undisplay <objectname> 

   Stops objectname from being displayed.  Nothing happens
   if object has not been displayed. In SMVR, all objects 
   are unique.  It is also the users responsibility to 
   keep track of locations (display by itself is not defined)

   >color <objectname> <r> <g> <b>

   causes <objectname> to have the colors specified by the
   r,g,b pairs when displayed.  r, g, and b are floats between
   [0,1]
	
--the following features may be unsupported as of current release

   >move <objectname> <x2> <y2> <z2> 
   >move <objectname> <x1> <y1> <z1> <x2> <y2> <z2>

   Displays an animated <objectname> moving from x1,y1,z1
   to x2,y2, z2.  This feature might not be implemented yet

   >scale <objectname> <sx> <sy> <sz>

   objects are initially scaled to 1 by 1 by 1
   this scales/distorts the x,y,z dimensions

   >rotate <objectname> <ro> <theta> <phi>

   Rotates object.  Not sure about behavior yet.

*/


/* for non-blocking IO */
#include<sys/time.h>
#include<sys/select.h> 

/* include all the good cave stuff */
#include <cave_ogl.h>
#include <GL/glu.h>
#include <GL/glut.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* aaargh! */
#include <curses.h>	/* for getch */

/* for importing WaveFront objects */

#include "glm.h"

/* for tokenizing strings */ 
#include "tokenize.h"


#include "smvr.h"


#define MAX_ARGS 20
#define MAX 100 
#define ERR -10000
#define HELP_TEXT "Simple Minded Virtual Reality\n\
commands:\n\
quit	--quits program\n\
load OBJECT FILENAME --loads wavefront file into OBJECT\n\
display OBJECT [x y z [sx sy sz [ro theta phi [r g b]]]]\n\
undisplay OBJECT\n\
locate OBJECT x y z\n\
color OBJECT r g b\n\
move OBJECT x y z --moves OBJECT to x y z\n"

typedef struct {

  char cmd[MAX];
  char object[MAX];
  float args[MAX_ARGS];
  int num_args;

}cmdstruct;

typedef enum {LIMBO, INIT, PLAY} state;  

/* globals */
 float redMaterial[] = { 1, 0, 0, 1 };
 float whiteMaterial[]={1,1,1,1};
 float blueMaterial[]={0,0,1,1};
 float greenMaterial[]={0,1,0,1};
 float Material[]={1,1,1,1};

int dlbase=0;	/* display list base */
int Gargc;
char** Gargv;
int Gcurrent_list=0;


int getcmd(char* in, cmdstruct *command);	/* 1 on success, 0 on fail */
void dumpcmd(cmdstruct* command);
void chomp(char* input, int input_size);


void init_gl(void);
void draw_objects(struct _object**);
void check_add(struct _object **list);
struct _object* update_held(struct _object *obj);
struct _object* check_grab(struct _object **list);

int main(int argc, char** argv){

  /* display related */
  struct _object **objList, *heldObj=NULL;
  float wandPos[3],wandFront[3];
  struct _object* obj=NULL;
  /* added by Oguz 3/7/1999 */
   GLMmodel* model;
    int i,j,k,current_list;
    float redMaterial[] = { 1, 0, 0, 1 };
    int ok=1;
   

  int done=0;
  char input[MAX]={'\0'};
  cmdstruct cmd;
  state s=LIMBO; 
  char** ltokens;	/* local tokens */
  int ltsize;		/* size of ltokens */

  /* for nonblocking IO */
  struct timeval tp;
  fd_set fds;
  char ch='x';
  char in_buffer[100]={'\0'};
  char dummy_buffer[100]={'\0'};
  int stdin_fd=fileno(stdin);
  tp.tv_sec = 0;
  tp.tv_usec = 0;
 

  Gargc=argc;
  Gargv=argv;

 CAVEConfigure(&argc,argv,NULL);
 /* Initialize the list pointer */
  objList = (struct _object **) CAVEMalloc(sizeof(struct _object *));
   *objList = NULL;

    CAVEInit();
    CAVEInitApplication(init_gl,0);
    CAVEDisplay(draw_objects,1,objList);
    
    dlbase=glGenLists(MAX);    





 printf("\nwelcome to SMVR by Oguz Yetkin,\
	 type help for assistance"); 
 printf("\nsmvr>");	/* print the first prompt */
 
  while(!done){
    FD_ZERO(&fds);
    FD_SET(stdin_fd, &fds);
    
  /* exit rudely if esc is pressed */
    if(CAVEgetbutton(CAVE_ESCKEY)){
     CAVEExit(); 
     done=1;
     return 1;
    }
    check_add(objList);		/* remnant from old program */
    if(heldObj)
	heldObj=update_held(heldObj);
    else
	heldObj=check_grab(objList); 
    sginap(1);


    /* the following block is non-blocking (no pun intended) */
    /* non-blockingness implemented 3/24/1999 */
 
   if( select(1, &fds, NULL, NULL, &tp) )    {
              gets(input);
	      /* chomp(input, MAX);  */


    /* scanf("%s\n",input); */
    /* getline(input); */
    /* fgets(input, MAX, stdin); */

   
 
    /* stateless operations */

    if(strncasecmp(input, "help", 4)==0){
	printf(HELP_TEXT);
    }else if(strncasecmp(input,"quit",4)==0){
	done=1;
    }

    switch(s){ 			/*switch on state */
	case LIMBO:
	 if(strncasecmp(input, "init",4)==0){
	   s=INIT;
	 }else{
	  printf("\n-ERR issue init command\n");
	 }
	break;

	case INIT:
	 ltsize=tokenize(&ltokens, input);
	 
	 /* load the object */
	 if(strncasecmp(ltokens[0], "load", 4)==0){
	  printf("\n^loading file %s into %s",ltokens[2],ltokens[1]);
	  obj = (struct _object *) CAVEMalloc(sizeof(struct _object)); 
	  if (!obj)
                {
                fprintf(stderr,"OUT OF MEMORY!\n");
                return 1;
         }
	 obj->x=0;
	 obj->y=0;
	 obj->z=0;
	 obj->r=0.8;
	 obj->g=0.8;
	 obj->b=1;
	 obj->sx=1;
	 obj->sy=1;
	 obj->sz=1;
	 obj->ro=0;
	 obj->theta=0;
	 obj->phi=0;
	 obj->radius=1;
	 obj->grabbed=0;
	 strcpy(obj->filename, ltokens[2]);
	 strcpy(obj->name,ltokens[1]);

	 /* now add object to list */
	 obj->next = *objList;
	 *objList = obj;




	}else if(strncasecmp(ltokens[0], "endinit", 4)==0){
	  s=PLAY;
	  /* now go through all the objects and actually render 
	     them */
	  Gcurrent_list=dlbase;
	 
	  /* iterate through list, render objects */
	/*	 
	 glEnable(GL_LIGHTING);
	  glEnable(GL_LIGHT0); 
	*/
	 /* material commented out 3/25/1999 */
	   glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, whiteMaterial); 
	  for (obj= *objList; obj; obj= obj->next){
	   Gcurrent_list++;
	/* we don't need to load models now, only need to display 
	model=glmReadOBJ(obj->filename);
	   glmUnitize(model);
	   glNewList(Gcurrent_list, GL_COMPILE);
	    glmDraw(model, GLM_TEXTURE|GLM_MATERIAL|GLM_SMOOTH);
	   glEndList();
	  printf("\nloaded file %s into object %s Gcurrent_list is: %d",obj->filename, obj->name,Gcurrent_list);
	*/
	  obj->list_id=Gcurrent_list;
	  obj->display=0;
	  obj->rendered=0;
	  
          }/* end for */
	 } 
	 free2(ltokens,ltsize);
	
	break;

	case PLAY:
         ok=getcmd(input, &cmd);
	 if(!ok){
		printf("\nnot OK, we should skip all!\n");
	}
         /* dumpcmd(&cmd); */

         if(ok && strncasecmp(cmd.cmd,"display",4)==0){
        	switch(cmd.num_args){
		 case 0:
		    obj=smvrfind(objList, cmd.object);
		    if(obj){ 
		     obj->display=1;
		     printf("\ndisplaying %s at %f %f %f",
			obj->name, obj->x, obj->y, obj->z);
		    }else{
		     printf("\n-ERR object not found");
		    }
		 break; 
  		 case 3: 	 
		   obj=smvrfind(objList, cmd.object);
		   if(obj){
		    obj->x=cmd.args[2];
		    obj->y=cmd.args[3];
		    obj->z=cmd.args[4];
		    obj->display=1;
		    printf("\ndisplaying %s at %f %f %f",
			   obj->name, obj->x, obj->y, obj->z);
		   }else{
		     printf("\n-ERR object not found");
		   }


		 break;

		 case 6:


		 break;

		 case 9:


		 break;


		 case 12:

		 break;

		 default:
		  printf("\n-ERR improper num of args for display\n");


		} /* end switch */
         }else
	 if(ok && strncasecmp(cmd.cmd,"undisplay",4)==0){
	        if(cmd.num_args<=0){	
		 printf("\n^undisplaying %s\n",cmd.object);
		obj=smvrfind(objList, cmd.object);
		 if(obj){
		  obj->display=0;
		 }else{
		  printf("\n-ERR object not found"); 
		 }

		}else{
		 printf("\n-ERR improper number of args for undisplay");
	  	}	

	 }else
	 if(ok && strncasecmp(cmd.cmd,"locate",4)==0){
		 obj=smvrfind(objList, cmd.object);
 		 switch(cmd.num_args){
                 case 3: 
		  if(obj){
                    obj->x=cmd.args[2];
                    obj->y=cmd.args[3];
                    obj->z=cmd.args[4];
                   }else{
                     printf("\n-ERR object not found");
                   }
		 break;
		 default:
		   printf("\n-ERR incorrect number of arguments for locate");
		break;	
	       } /* end switch */
	 }else
	 if(ok && strncasecmp(cmd.cmd,"color",4)==0){
		 obj=smvrfind(objList, cmd.object);
                 switch(cmd.num_args){
                 case 3:
                  if(obj){
                    obj->r=cmd.args[2];
                    obj->g=cmd.args[3];
                    obj->b=cmd.args[4];
                   }else{
                     printf("\n-ERR object not found");
                   }
                 break;
                 default:
                   printf("\n-ERR incorrect number of arguments for color");
                break;
               } /* end switch */



	 }else
	 if(ok && strncasecmp(cmd.cmd,"rotate",4)==0){
 		obj=smvrfind(objList, cmd.object);
                 switch(cmd.num_args){
                 case 3:
                  if(obj){
                    obj->ro=cmd.args[2];
                    obj->theta=cmd.args[3];
                    obj->phi=cmd.args[4];
                   }else{
                     printf("\n-ERR object not found");
                   }
                 break;
                 default:
                   printf("\n-ERR incorrect number of arguments for rotate");
                break;
               } /* end switch */

	 }else
	 if(ok && strncasecmp(cmd.cmd,"scale",4)==0){
	   obj=smvrfind(objList, cmd.object);
                 switch(cmd.num_args){
                 case 3:
                  if(obj){
                    obj->sx=cmd.args[2];
                    obj->sy=cmd.args[3];
                    obj->sz=cmd.args[4];
                   }else{
                     printf("\n-ERR object not found");
                   }
                 break;
                 default:
                   printf("\n-ERR incorrect number of arguments for scale");
                break;
               } /* end switch */



	 }else
	 if(ok && strncasecmp(cmd.cmd,"move",4)==0){
	 	switch(cmd.num_args){
		 case 3:
		  
		 break;

		 case 6:

		 break;

		} /* end switch */
	 }


 
 	break; 			/* end state PLAY */
    } /* end outer switch */
    printf("\nsmvr>");
   }  /* end non-blocking block */

   } /* end outer while */ 



  return 0;
}


struct _object* smvrfind(struct _object** objList, char* str){

 struct _object* obj;
 for (obj= *objList; obj; obj= obj->next){
  if(strcmp(obj->name, str)==0){
   printf("\n^smvrfind found %s, returning it\n", str);
   return obj;
  }
 }
 return NULL;


}


int getcmd(char* in, cmdstruct *command){

	char** tokens;
  	int num_tokens;
	int i=0;

	num_tokens=tokenize(&tokens, in);
	if(num_tokens==0){
		printf("\nerror: no tokens");
		return 0;
	}
	command->num_args=num_tokens-2;	
	strcpy(command->cmd, tokens[0]); 
	strcpy(command->object, tokens[1]);
	for(i=2;i<MAX_ARGS;i++){	/* clear of extra stuff */
		(command->args)[i]=ERR;
	}

	for(i=2;i<num_tokens;i++){
		(command->args)[i]=(float)strtod(tokens[i],NULL);
	}
	
	/* free the memory we alloc'd */
	free2(tokens, num_tokens);
 	return 1;
}

/* chomp() removes carriage returns */
void chomp(char* input, int input_size){
 int i=0;
 for(i=0;i<input_size;i++){
        if(input[i]=='\n'){ 
         input[i]='\0';
        }
 }   
}



/* dumpcmd
   for debugging
*/
void dumpcmd(cmdstruct* command){
	int i=0;

	printf("\ncommand: %s",command->cmd);
	printf("\nobject: %s",command->object);
	printf("\nnum_args: %d args:\n", command->num_args);
	/* commented out 3/24/1999 */
  	
	for(i=2;(command->args)[i]!=ERR && i<MAX_ARGS;i++){
		printf("%f ",(command->args)[i]);
	}
	printf("\n");


}


void init_gl(void){
 /* added by Oguz 3/7/1999 */
 /* for smvr, we can't glmDraw anything here */

 GLMmodel* model;
/* 
 float redMaterial[] = { 1, 0, 0, 1 };
 float whiteMaterial[]={1,1,1,1};
 float blueMaterial[]={0,0,1,1};
 float greenMaterial[]={0,1,0,1};
*/
 int i,j,k,current_list;
  glEnable(GL_LIGHT0);
  glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, whiteMaterial); 
 /* sphereObj = gluNewQuadric(); */
/*
 dlbase=glGenLists(Gargc);
*/

 printf("\ndlbase in init_gl is: %d\n",dlbase);

  current_list=dlbase;

/* commented out 3/24/1999

 for(i=1;i<Gargc;i++){
   Gcurrent_list++;  
   model=glmReadOBJ(Gargv[i]);
   glmUnitize(model);
   glNewList(Gcurrent_list, GL_COMPILE);
   glmDraw(model, GLM_TEXTURE|GLM_MATERIAL|GLM_SMOOTH);
   glEndList();
   printf("\nloaded object %s current list is %d",Gargv[i],current_list);
   current_list++;
*/
}

void draw_objects(struct _object** objList){
 struct _object *obj;
 GLMmodel* model;
 float wandAngle[3]; /* x y z, ro theta phi, or elevation azimuth roll */
 glClearColor(0., 0., 0., 0.);
 glClear(GL_DEPTH_BUFFER_BIT|GL_COLOR_BUFFER_BIT);
 glEnable(GL_LIGHTING);
        /* draw a 'cursor' by OY 2/6/1999 */
        glPushMatrix();
        CAVEWandTransform();
        /* gluSphere(sphereObj, 0.5, 6,6); */
        /* glCallList(dlbase+1);  */
        /* commented out 2/16/1999 */
        glTranslatef(0.0,0,0.1);
        glRotatef(180,0,1,0);
        glutSolidCone(0.3, 0.7, 4, 3);
        glPopMatrix();

 for (obj= *objList; obj; obj= obj->next)
       /* extra check added 3/24/1999 for smvr */
       if(obj->display){ 
	if (obj->grabbed)
                {
                glPushMatrix();
                 CAVEWandTransform();
                 /* gluSphere(sphereObj, obj->radius, 8, 8); */
                glEnable(GL_NORMALIZE);
                 glScalef(obj->radius, obj->radius, obj->radius);
                glCallList(obj->list_id);
                glDisable(GL_NORMALIZE);
                glPopMatrix();
		/* read the orientation of wand and record it */
		CAVEGetOrientation(CAVE_WAND, wandAngle);
		obj->ro=wandAngle[0];
		obj->theta=wandAngle[1];
		obj->phi=wandAngle[2];	
                }
        else
                {
                glPushMatrix();
                glTranslatef(obj->x, obj->y, obj->z);
		/* glRotatef(obj->ro, obj->theta, obj->phi,0); */
		/* rotation to be optimized later, 3/28/1999 */
 		glPushMatrix(); /* for rotate */
		 glRotatef(obj->ro, 1, 0, 0);
		 glRotatef(obj->theta, 0, 1, 0);
		 glRotatef(obj->phi, 0, 0, 1);
                /* gluSphere(sphereObj, obj->radius, 8, 8); */
                glEnable(GL_NORMALIZE);
                glScalef(obj->radius, obj->radius, obj->radius);
                /* glCallList(obj->list_id); */
		/* glutSolidCube(0.5); */
		if(!(obj->rendered)){
		 glNewList(obj->list_id,GL_COMPILE); 
		  model=model=glmReadOBJ(obj->filename);
           	  glmUnitize(model);
		  glmDraw(model, GLM_TEXTURE|GLM_MATERIAL|GLM_SMOOTH);
		 glEndList();
		 obj->rendered=1;
	        }
		glScalef(obj->sx, obj->sy, obj->sz);
		glColor3f(obj->r, obj->g, obj->b);
		Material[0]=obj->r;
		Material[1]=obj->g;
		Material[2]=obj->b;
		glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, Material);
		glCallList(obj->list_id);
		glPopMatrix(); /* for rotate */
                glDisable(GL_NORMALIZE);
                glPopMatrix();
		/*	
		printf("\ndisplayed %s from file %s at %f %f %f \
			 with list_id %d",
		     obj->name, obj->filename, obj->x, obj->y, obj->z,
		     obj->list_id);
		*/
                }
	} /* end outer if */
}


/* possibly not too useful for smvr, we'll just add spheres
   to verify that the program is working */
void check_add(struct _object **list){

 if (CAVEButtonChange(3) == -1)
        {
        struct _object *obj;
        float wandPos[3],wandFront[3];
        obj = (struct _object *) CAVEMalloc(sizeof(struct _object));
        if (!obj)
                {
                fprintf(stderr,"OUT OF MEMORY!\n");
                return;
                }
        CAVEGetPosition(CAVE_WAND,wandPos);
        CAVEGetVector(CAVE_WAND_FRONT,wandFront);
        obj->x = wandPos[0] + wandFront[0]*2;
        obj->y = wandPos[1] + wandFront[1]*2;
        obj->z = wandPos[2] + wandFront[2]*2;
        obj->radius = 1;
        obj->grabbed = 0;
        printf("\ndlbase is now: %d",dlbase);
	/* changed for smvr */
        obj->list_id=1+dlbase;
	printf("\nassigned list id: %d", obj->list_id);
	
        obj->next = *list;
        *list = obj;
        }
	



}
struct _object* update_held(struct _object *obj){

if (!CAVEBUTTON1)
        {
        float wandPos[3];
        CAVEGetPosition(CAVE_WAND,wandPos);
        obj->x = wandPos[0];
        obj->y = wandPos[1];
        obj->z = wandPos[2];
        obj->grabbed = 0;
        return NULL;
        }
 if (CAVE_JOYSTICK_Y > .5)
        obj->radius *= 1.005;
 else if (CAVE_JOYSTICK_Y < -.5)
        obj->radius /= 1.005;
 return obj;

}

struct _object* check_grab(struct _object **list){
 if (CAVEButtonChange(1) == 1)
        {
        struct _object *obj;
        float wandPos[3];
        CAVEGetPosition(CAVE_WAND,wandPos);
        for (obj = *list; obj; obj = obj->next)
                {
                float dist;
#define SQR(x) ((x)*(x))
                dist = SQR(obj->x - wandPos[0]) + SQR(obj->y - wandPos[1]) +
                        SQR(obj->z - wandPos[2]);
                if (dist < SQR(obj->radius))
                        {
                        obj->grabbed = 1;
                        return obj;
                        }
                }
        }
 return NULL;

}
