/* interact.c
/*    An example of an interactive CAVE environment. The user is able to
/*   add spheres to the environment using the right wand button, to grab
/*   existing spheres using the left button, and to resize a grabbed sphere
/*   with the joystick. The data for the spheres are kept in a linked list
/*   in shared memory, and are updated by the main computation process.
*/

#include <cave_ogl.h>
#include <GL/glu.h>

/* global variable, added by OY 2/6/1999 */

int currentlist=0;
static GLUquadricObj *sphereObj;

/* The data that will be shared between processes */
struct _balldata
	{
	float x,y,z;
	float radius;
	int grabbed;
	float r,g,b;
	int displaylist;
	struct _balldata *next;
	};

void init_gl(void),draw_balls(struct _balldata **);
void check_add(struct _balldata **list);
struct _balldata * update_held(struct _balldata *ball);
struct _balldata * check_grab(struct _balldata **list);

main(int argc,char **argv)
{
 struct _balldata **ballList,*heldBall=NULL;
 CAVEConfigure(&argc,argv,NULL);
/* Initialize the list pointer */
 ballList = (struct _balldata **) CAVEMalloc(sizeof(struct _balldata *));
 *ballList = NULL;
 CAVEInit();
 CAVEInitApplication(init_gl,0);
 CAVEDisplay(draw_balls,1,ballList);
 while (!CAVEgetbutton(CAVE_ESCKEY))
	{
	/* added by OY 2/6/1999 */

	glPushMatrix();
        CAVEWandTransform();	
	glColor3f(0,0,1);
        gluSphere(sphereObj, 0.5, 5, 5);
	

	glPopMatrix();
	check_add(ballList);
	if (heldBall)
		heldBall = update_held(heldBall);
	else
		heldBall = check_grab(ballList);
	sginap(1);
	}
 CAVEExit();
}


void check_add(struct _balldata **list)
{
 if (CAVEButtonChange(3) == -1)
	{
	struct _balldata *ball;
	float wandPos[3],wandFront[3];
	ball = (struct _balldata *) CAVEMalloc(sizeof(struct _balldata));
	/* apparently allocates memory for the ball struct OY 2/6/1999 */
	if (!ball)
		{
		fprintf(stderr,"OUT OF MEMORY!\n");
		return;
		}
	CAVEGetPosition(CAVE_WAND,wandPos);
	CAVEGetVector(CAVE_WAND_FRONT,wandFront);
	ball->x = wandPos[0] + wandFront[0]*2;
	ball->y = wandPos[1] + wandFront[1]*2;
	ball->z = wandPos[2] + wandFront[2]*2;
	ball->r=1;
	ball->g=0;
	ball->b=0;
	ball->displaylist=currentlist++;
	glNewList(ball->displaylist, GL_COMPILE);
	  glColor3f(ball->r, ball->g, ball->b);		
	  gluSphere(sphereObj, ball->radius, 8, 8);
	glEndList();	
			
	ball->radius = 1;
	ball->grabbed = 0;
	ball->next = *list;
	*list = ball;
	}
}


struct _balldata * update_held(struct _balldata *ball)
{
 if (!CAVEBUTTON1)
	{
	float wandPos[3];
	CAVEGetPosition(CAVE_WAND,wandPos);
	ball->x = wandPos[0];
	ball->y = wandPos[1];
	ball->z = wandPos[2];
	ball->grabbed = 0;
	return NULL;
	}
 if (CAVE_JOYSTICK_Y > .5)
	ball->radius *= 1.005;
 else if (CAVE_JOYSTICK_Y < -.5)
	ball->radius /= 1.005;
 return ball;
}


struct _balldata * check_grab(struct _balldata **list)
{
 if (CAVEButtonChange(1) == 1)
	{
	struct _balldata *ball;
	float wandPos[3];
	CAVEGetPosition(CAVE_WAND,wandPos);
	for (ball = *list; ball; ball = ball->next)
		{
		float dist;
#define SQR(x) ((x)*(x))
		dist = SQR(ball->x - wandPos[0]) + SQR(ball->y - wandPos[1]) + 
			SQR(ball->z - wandPos[2]);
		if (dist < SQR(ball->radius))
			{
			ball->grabbed = 1;
			return ball;
			}
		}
	}
	/* check grab for color change */
	/* added by Oguz */
 if (CAVEButtonChange(2) == 1)
        {
        struct _balldata *ball;
        float wandPos[3];
        CAVEGetPosition(CAVE_WAND,wandPos);
        for (ball = *list; ball; ball = ball->next)
                {
                float dist;
#define SQR(x) ((x)*(x))
                dist = SQR(ball->x - wandPos[0]) + SQR(ball->y - wandPos[1]) +
                        SQR(ball->z - wandPos[2]);
                if (dist < SQR(ball->radius))
                        {
                        /* ball->grabbed = 0; */
			ball->r=rand();
			ball->g=rand()/ball->y;
			ball->b=rand()/ball->x;
                        return ball;
                        }
                }
        }



 return NULL;
}





/* init_gl - initialize GL lighting & materials */
void init_gl(void)
{
 float redMaterial[] = { 1, 0, 0, 1 };
 glEnable(GL_LIGHT0);
 glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, redMaterial);
 sphereObj = gluNewQuadric();
}


/* draw_balls - draw the balls, using the shared data for their positions
  & sizes */
void draw_balls(struct _balldata **ballList)
{
 struct _balldata *ball;

 glClearColor(0., 0., 0., 0.);
 glClear(GL_DEPTH_BUFFER_BIT|GL_COLOR_BUFFER_BIT);
 glEnable(GL_LIGHTING|GL_COLOR);

for (ball = *ballList; ball; ball = ball->next)
	if (ball->grabbed)
		{
		glPushMatrix();
		 CAVEWandTransform();
		/* added by Oguz */		
		/* glColor3f(0, 1, 0); */
		 /* gluSphere(sphereObj, ball->radius, 8, 8); */
		glCallList(ball->displaylist);
		glPopMatrix();
		}
	else
		{
		glPushMatrix();
		 glTranslatef(ball->x, ball->y, ball->z);
		/*	
		glColor3f(ball->r,ball->g,ball->b);
		 gluSphere(sphereObj, ball->radius, 8, 8);
		*/
		glCallList(ball->displaylist);
		glPopMatrix();
		}

 glDisable(GL_LIGHTING);
}
