/* interact.c
/*    An example of an interactive CAVE environment. The user is able to
/*   add spheres to the environment using the right wand button, to grab
/*   existing spheres using the left button, and to resize a grabbed sphere
/*   with the joystick. The data for the spheres are kept in a linked list
/*   in shared memory, and are updated by the main computation process.
*/

#include <cave_ogl.h>
#include <GL/glu.h>
#include <GL/glut.h>

static GLUquadricObj *sphereObj;
int dlbase=0;	/* display list base */
/* The data that will be shared between processes */
struct _balldata
	{
	float x,y,z;
	float radius;
	int grabbed;
	struct _balldata *next;
	};

void init_gl(void),draw_balls(struct _balldata **);
void check_add(struct _balldata **list);
struct _balldata * update_held(struct _balldata *ball);
struct _balldata * check_grab(struct _balldata **list);

main(int argc,char **argv)
{
 struct _balldata **ballList,*heldBall=NULL;
 float wandPos[3],wandFront[3];
 CAVEConfigure(&argc,argv,NULL);
/* Initialize the list pointer */
 ballList = (struct _balldata **) CAVEMalloc(sizeof(struct _balldata *));
 *ballList = NULL;
 CAVEInit();
 CAVEInitApplication(init_gl,0);
 CAVEDisplay(draw_balls,1,ballList);
 while (!CAVEgetbutton(CAVE_ESCKEY))
	{
	/* end block added by OY to get cursor */

	check_add(ballList);
	if (heldBall)
		heldBall = update_held(heldBall);
	else
		heldBall = check_grab(ballList);
	sginap(1);

		
	}
 CAVEExit();
}


void check_add(struct _balldata **list)
{
 if (CAVEButtonChange(3) == -1)
	{
	struct _balldata *ball;
	float wandPos[3],wandFront[3];
	ball = (struct _balldata *) CAVEMalloc(sizeof(struct _balldata));
	if (!ball)
		{
		fprintf(stderr,"OUT OF MEMORY!\n");
		return;
		}
	CAVEGetPosition(CAVE_WAND,wandPos);
	CAVEGetVector(CAVE_WAND_FRONT,wandFront);
	/* changed 4/2/1999 
	ball->x = wandPos[0] + wandFront[0]*2;
	ball->y = wandPos[1] + wandFront[1]*2;
	ball->z = wandPos[2] + wandFront[2]*2;
	*/
	ball->x = wandPos[0];
        ball->y = wandPos[1];
        ball->z = wandPos[2];

	ball->radius = 0.2;
	ball->grabbed = 0;
	ball->next = *list;
	*list = ball;
	/* added by OY 3/28/1999 to determine good locations */
	printf("\nx y z: %f %f %f",ball->x, ball->y, ball->z);
	}
}


struct _balldata * update_held(struct _balldata *ball)
{
 if (!CAVEBUTTON1)
	{
	float wandPos[3];
	CAVEGetPosition(CAVE_WAND,wandPos);
	ball->x = wandPos[0];
	ball->y = wandPos[1];
	ball->z = wandPos[2];
	ball->grabbed = 0;
	return NULL;
	}
 if (CAVE_JOYSTICK_Y > .5)
	ball->radius *= 1.005;
 else if (CAVE_JOYSTICK_Y < -.5)
	ball->radius /= 1.005;
 return ball;
}


struct _balldata * check_grab(struct _balldata **list)
{
 if (CAVEButtonChange(1) == 1)
	{
	struct _balldata *ball;
	float wandPos[3];
	CAVEGetPosition(CAVE_WAND,wandPos);
	for (ball = *list; ball; ball = ball->next)
		{
		float dist;
#define SQR(x) ((x)*(x))
		dist = SQR(ball->x - wandPos[0]) + SQR(ball->y - wandPos[1]) + 
			SQR(ball->z - wandPos[2]);
		if (dist < SQR(ball->radius))
			{
			ball->grabbed = 1;
			return ball;
			}
		}
	}
 return NULL;
}





/* init_gl - initialize GL lighting & materials */
void init_gl(void)
{
 float redMaterial[] = { 1, 0, 0, 1 };
 glEnable(GL_LIGHT0);
 glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, redMaterial);
 sphereObj = gluNewQuadric();
 dlbase=glGenLists(10);	/* generate 10 empty lists, OY 1/6/1999 */
 glNewList(dlbase, GL_COMPILE);
	
	/* gluSphere(sphereObj, 0.5, 6,6); */
	glBegin(GL_POLYGON); 
	glColor3f(0,0,1);
	glVertex3f(0,0,0);
	glVertex3f(0,0.5,0);
	glVertex3f(0,0.5, 0.5);
	glVertex3f(0.2,0.4, 0.1);
	glEnd();
 glEndList();	

}


/* draw_balls - draw the balls, using the shared data for their positions
  & sizes */
void draw_balls(struct _balldata **ballList)
{
 struct _balldata *ball;

 glClearColor(0., 0., 0., 0.);
 glClear(GL_DEPTH_BUFFER_BIT|GL_COLOR_BUFFER_BIT);
 glEnable(GL_LIGHTING);
	/* draw a 'cursor' by OY 2/6/1999 */
	glPushMatrix();
	CAVEWandTransform();
	/* gluSphere(sphereObj, 0.5, 6,6); */
	/* glCallList(dlbase); */
	/* commented out 2/16/1999 */
	glTranslatef(0.0,0,0.1);
	glRotatef(180,0,1,0);	
	glutSolidCone(0.3, 0.7, 4, 3);


	glPopMatrix();
 for (ball = *ballList; ball; ball = ball->next)
	if (ball->grabbed)
		{
		glPushMatrix();
		 CAVEWandTransform();
		 gluSphere(sphereObj, ball->radius, 8, 8);
		glPopMatrix();
		}
	else
		{
		glPushMatrix();
		 glTranslatef(ball->x, ball->y, ball->z);
		 gluSphere(sphereObj, ball->radius, 8, 8);
		glPopMatrix();
		}

 glDisable(GL_LIGHTING);
}
