/* $id: smvr.c,v 1.12.2.4 1999/07/01 04:06:21 caswick Exp $ */
/* smvr.c 3/14/1999 Oguz Yetkin (oguz@writeme.com 
   Simple Minded Virtual Reality (SMVR) server
   Working Version: 3/28/1999 2:36 am
   Copy Implemented: 4/5/1999 2:47 am
   Write Implemented: 4/5/1999, sometime around 2 am
   Display Arguments completely implemented:  4/5/1999 2:47 am 
   Navigation (almost) Properly Implemented: 4/6/1999 12:14 am 
   Modified OY 5/8/1999 to fix rotations
   Modified OY 6/10/1999 to add text output
   Modified OY 6/13/1999 to add navigation while holding object  
	(grab object with middle button and navigate with joystick,
	 right now (4:26 pm), the 3-D looks messed up during carry)  
*/
/*
$Log: smvr.c,v $
Revision 1.16  1999/07/24 19:19:50  caswick
Removed some old cruft
Made it so it only unpauses on button press, so that multiple pauses don't release when you hold the button

Revision 1.15  1999/07/24 00:59:04  caswick
Rearranged ifs so that it doesn't go into the switch statement if one of the stateless commands executes
Prevents oddball crashes

Revision 1.14  1999/07/24 00:51:17  caswick
Made getcmd() not strcpy from invalid pointers when there are less than 3 tokens

Revision 1.13  1999/07/19 23:59:12  caswick
Just merged in RedmondDemo-1999Jun branch, haven't verified compile yet

Revision 1.12.2.5  1999/07/01 05:33:41  caswick
Adjusted alignment of text to be better centered on display

Revision 1.12.2.4  1999/07/01 04:06:21  caswick
Made pause, unpause, ispaused commands be stateless

Revision 1.12.2.3  1999/07/01 02:59:14  caswick
Changed MAX from 100 to 500 so that the credits message can be printed

Revision 1.12.2.2  1999/07/01 02:44:50  caswick
I was right, didn't need the #if 0s, the print buffer overflow was messing with the flags
So I removed them

Revision 1.12.2.1  1999/07/01 02:38:49  caswick
#if 0'd out the physics and funky code (though that might not be necessary anymore)
Added test to print command for buffer overflow
Removed redundant test of pause command

Revision 1.12  1999/06/29 08:18:12  yetkin
Newtonian physics pretty much implemented
added physicson, gravityon, physicsoff, gravityoff
gravity is a little too strong right now, but works properly.
phsics needs to be fine tuned a little bit.
Next step: buoyancy, air density, volume, etc.  Also, bound checking (only ground is checked now)

Revision 1.11  1999/06/29 05:41:36  yetkin
funkyon and funkyoff commands now work

Revision 1.10  1999/06/29 02:23:58  yetkin
throw now works!!!

Revision 1.9  1999/06/29 01:24:59  yetkin
implemented pause, ispaused, and unpause.  User can now release "pause" with right
wand button.

Revision 1.8  1999/06/24 06:39:21  yetkin
intermediate version: fixed scaling problem, reduced rotation speed,
added globals Gss->Gvx Gss->Gvy and Gss->Gvz containing
the velocity of the wand.  read_wand() updates these quantities.
Also, Gss->Gfunky enables funky special effects (mostly for testing velocity.
Next version: smvrfind() (clone) fix and simple physics

Revision 1.7  1999/06/13 21:15:50  yetkin
Added carry (use joystick to navigate while object is held with middle button).
Right now, the 3-D looks somewhat messed up during the carry

Revision 1.6  1999/06/11 21:39:56  yetkin
Worked more on text, now you can use '|' as a return character, works better

Revision 1.5  1999/06/11 20:26:40  yetkin
Added "print" function for text output

 * Revision 1.4  1999/04/24  20:02:10  yetkin
 * Moved holodeck walls into a function
 *
 * Revision 1.3  1999/04/24  19:53:24  yetkin
 * added ability to toggle walls, spheres, and motion speed with function keys
 *
 * Revision 1.2  1999/04/24  19:38:35  caswick
 * Added $Id and $Log to file
 *
*/  
/*  future goal: color palette (I suppose this is turning into a CAD program)
    practical goal: inventory function

   Cave Version
   
   WARNING: Keep this file in pure ANSI C for Cave compatibility
	    Do not pollute with C++.
	    (this library was written to separate the libraries
	     from C++ in the first place)

   responds to text requests, loading Wavefront objects
   assigning them names, translating, scaling, and displaying.

   This program initially written for CS774 to separate
   object loading and display from the program logic.

   The following operations are legal within their blocks.
   The blocks must be executed in order.

   1) Initialization State
   
   #>objects <num>  
  
   #<num> is an integer specifying number of maximum 
   #objects for this session 
      
   >init 
  
   >load <objectname> <filename>
   
   load can be repeated <num> times.  
   Loads an object from a wavefront .obj file
   load will overwrite previous definition if it exists.
   >endinit

   2) Display State 
  
    >locate <objectname> <x> <y> <z>
    
    >display <objectname> #defaults to origin 
    >display <objectname> <x> <y> <z> 
    >display <objectname> <x> <y> <z> <sx> <sy> <sz> <ro> <theta> <phi> 
    >display <objectname> <x> <y> <z> <sx> <sy> <sz> <ro> <theta> <phi> <r> <g> <b>
 
   where <objectname> is the string assigned to the object
   in block 2, and <x> <y> <z> are coordinates of the object
   In the cave, x E [-5,5] y E [0,10] z E [-5,5].  Positive z is
   towards the user, x is horizontal, and y is vertical

   >undisplay <objectname> 

   Stops objectname from being displayed.  Nothing happens
   if object has not been displayed. In SMVR, all objects 
   are unique.  It is also the users responsibility to 
   keep track of locations (display by itself is not defined)

   >color <objectname> <r> <g> <b>

   causes <objectname> to have the colors specified by the
   r,g,b pairs when displayed.  r, g, and b are floats between
   [0,1]

   >write <filename>

   writes a script file that includes the commands to generate
   the current scene.  By default, the current scene is saved
   under session.smvr.  Note that the WaveFront .obj 
   files used to create the scene must still be present 

--the following features may be unsupported as of current release

   >move <objectname> <x2> <y2> <z2> 
   >move <objectname> <x1> <y1> <z1> <x2> <y2> <z2>

   Displays an animated <objectname> moving from x1,y1,z1
   to x2,y2, z2.  This feature might not be implemented yet

   >scale <objectname> <sx> <sy> <sz>

   objects are initially scaled to 1 by 1 by 1
   this scales/distorts the x,y,z dimensions

   >rotate <objectname> <ro> <theta> <phi>

   Rotates object.  Not sure about behavior yet.

*/

static char rcsid[] = "$Id: smvr.c,v 1.16 1999/07/24 19:19:50 caswick Exp $";

#include<errno.h>
/* for non-blocking IO */
#include<sys/time.h>
#include<sys/select.h> 
/* include all the good cave stuff */
#include <cave_ogl.h>
#include <GL/glu.h>
#include <GL/glut.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* aaargh! */
#include <curses.h>	/* for getch */

/* for importing WaveFront objects */

#include "glm.h"

/* for tokenizing strings */ 
#include "tokenize.h"


#include "smvr.h"

/* for debug */ 
/* #define CAVENavTransform() */

#define MAX_ARGS 20
#define MAX 500 
#define ERR -10000
#define DEFAULT_FILE "session.smvr"
#define GRAV 32.0
#define PI 3.14
#define HELP_TEXT "Simple Minded Virtual Reality\n\
commands:\n\
quit	--quits program\n\
load OBJECT FILENAME --loads wavefront file into OBJECT\n\
display OBJECT [x y z [sx sy sz [ro theta phi [r g b]]]]\n\
undisplay OBJECT\n\
locate OBJECT x y z\n\
color OBJECT r g b\n\
move OBJECT x y z --moves OBJECT to x y z\n\
print text_to_be_printed\n\
inv OBJECT\n\
invall\n"

typedef enum {NORMAL=0, GRID, BRUSH, ERASE} togglestate;

typedef struct {

  char cmd[MAX];
  char object[MAX];
  char object2[MAX];
  float args[MAX_ARGS];
  int num_args;

}cmdstruct;

typedef enum {LIMBO, INIT, PLAY} state;  

/* globals */
 float redMaterial[] = { 1, 0, 0, 1 };
 float whiteMaterial[]={1,1,1,1};
 float blueMaterial[]={0,0,1,1};
 float greenMaterial[]={0,1,0,1};
 float Material[]={1,1,1,1};

int dlbase=0;	/* display list base */
int Gargc;
char** Gargv;
int Gcurrent_list=0;
int Gextra=0;	/* extra objects created by user */

int Gchanged=0;
float Gpro;	/* values of ro, theta, and phi
		   when an object is grabbed */ 
float Gptheta;
float Gpphi;	

/* added 6/23/1999 to measure wand velocity */

struct timeval Gptime;	/* time of last velocity change in seconds */
/* example: Gvx=(Gpx-Gx)/(Gptime-time) */
float Gpx;	/* previous x location */
float Gpy;
float Gpz;

/*
float Gvx;	  
float Gvy;
float Gvz;
*/


/* state variables */

typedef struct{ 
 int Gdraw_walls;	/* toggled by F1 */
 int Gdraw_bound;	/* toggled by F2 */
 float Gspeed;		/* decrease F3 increase F4 */
 float Grot_angle;	/* decrease F5 increase F6 */
 char Gtext_buffer[MAX]; /* OY 6/10/1999 to display text */
 float Gvx;	/* velocity */
 float Gvy;
 float Gvz;
 float Gdt;
 float Gdamping;   
 int Gfunky;
 int Gispaused;
 int Gphysics;	/* is there physics?  off by default */
 int Ggravity; /* is there gravity?  off by default (requires physics) */

}state_struct;

state_struct* Gss;

/* added 4/5/1999 to copy */
int Gcopy_object=0;	/* -1, Copy just done; 0, no, 1, yes */

void read_wand(void);	/* updates globals Gptime, Gvx, Gvy, Gvz */

int getcmd(char* in, cmdstruct *command);	/* 1 on success, 0 on fail */
void dumpcmd(cmdstruct* command);
void chomp(char* input, int input_size);
void navigate(void);
void draw_walls();

void init_gl(void);
void draw_objects(struct _object**);
void check_add(struct _object **list);

struct _object* update_held(struct _object *obj);
struct _object* check_grab(struct _object **list);
int smvrwrite(char* filename, struct _object **list); 
int toggle(int input);

float smvrmax(float a, float b, float c);

void smvrpollchanged(struct _object** objList);	/* KLC */

/* added 6/10/1999, renders a string of text */
void render(int list, char* string, int size);  /*OY */

void render2(char* string);


int main(int argc, char** argv){

  /* display related */
  struct _object **objList, *heldObj=NULL;
  float wandPos[3],wandFront[3];
  struct _object* obj=NULL;

  /* Cache button changes - added by KLC 7/24/1999 */
  int CBC1, CBC2, CBC3;

  /* added by Oguz 3/7/1999 */
   GLMmodel* model;
    int i,j,k,current_list;
    float redMaterial[] = { 1, 0, 0, 1 };
    int ok=1;
   

  int done=0;
  int opts=1;		/* command line options being processed */
  char input[MAX]={'\0'};
  cmdstruct cmd;
  state s=LIMBO; 
  char** ltokens;	/* local tokens */
  int ltsize;		/* size of ltokens */

  /* for nonblocking IO, bUNIX-SPECIFIC */
  struct timeval tp;
  fd_set fds;
  int stdin_fd;
  /* eUNIX-SPECIFIC */

  char ch='x';
  char in_buffer[100]={'\0'};
  char dummy_buffer[100]={'\0'};
  
  /* bUNIX-SPECIFIC */
  stdin_fd=fileno(stdin);
  tp.tv_sec = 0;
  tp.tv_usec = 0;
  /* eUNIX-SPECIFIC */


 

  Gargc=argc;
  Gargv=argv;
  
 
 CAVEConfigure(&argc,argv,NULL);
 /* Initialize the list pointer */
  objList = (struct _object **) CAVEMalloc(sizeof(struct _object *));
  
   *objList = NULL;


  Gss= (state_struct *) CAVEMalloc(sizeof(struct state_struct *));
  Gss->Gdraw_walls=1;
  Gss->Gdraw_bound=0;
  Gss->Gspeed=1.7;
  Gss->Grot_angle=45;
  Gss->Gfunky=0;
  Gss->Gispaused=0;
  Gss->Gphysics=0;
  Gss->Ggravity=0;
  Gss->Gdamping=0.99;
  Gss->Gdt=0.001;


    /* bCAVE-SPECIFIC */
    CAVEInit();
    CAVEInitApplication(init_gl,0);
    CAVEDisplay(draw_objects,1,objList);
    /* eCAVE-SPECIFIC */
 
    dlbase=glGenLists(MAX+1);    





 /* printf("\nwelcome to SMVR by Oguz Yetkin,\
	 type help for assistance");  */
 printf("\nsmvr>");	/* print the first prompt */
 
  while(!done){
goback:
    navigate(); 

    /* bUNIX-SPECIFIC */
    FD_ZERO(&fds);
    FD_SET(stdin_fd, &fds);
    /* eUNIX-SPECIFIC */
 
  /* exit rudely if esc is pressed */
    if(CAVEgetbutton(CAVE_ESCKEY)){
     smvrwrite(DEFAULT_FILE, objList);
     printf("\nscene written to %s\n", DEFAULT_FILE);
     printf("\n^QUIT\n" );

     CAVEExit(); 
     done=1;
     return 1;
    }
   /* printf("\nGss->Gdraw_walls in main: %d", Gss->Gdraw_walls); */
   if(CAVEgetbutton(CAVE_F1KEY)){
     Gss->Gdraw_walls=toggle(Gss->Gdraw_walls);
	/* printf("\ntoggling walls %d",Gss->Gdraw_walls); */
   }
   if(CAVEgetbutton(CAVE_F2KEY)){
     Gss->Gdraw_bound=toggle(Gss->Gdraw_bound);
   }
   if(CAVEgetbutton(CAVE_F3KEY)){
     Gss->Gspeed/=1.1;
    }
   if(CAVEgetbutton(CAVE_F4KEY)){
     Gss->Gspeed*=1.1;
    }
    if(CAVEgetbutton(CAVE_F5KEY)){
     Gss->Grot_angle--;
    } 
    if(CAVEgetbutton(CAVE_F6KEY)){
     Gss->Grot_angle++;
    }
	
	/* added 6/23/1999 */
    read_wand(); 
    check_add(objList);		/* remnant from old program */

    /* Read and cache button changes - KLC 7/24/1999 */
    CBC1 = CAVEButtonChange(1);
    CBC2 = CAVEButtonChange(2);
    CBC3 = CAVEButtonChange(3);

    if(heldObj){
      heldObj=update_held(heldObj);
    } else{
      if (CBC1 == 1 || CBC2 == 1 || CBC3 == 1)
	heldObj=check_grab(objList); 
      if(!heldObj){
	if( CBC3 == 1 ){ /* Check for button press instead of hold, so single unpause - KLC */
	  Gss->Gispaused=0;
	  printf("\n!released pause!");
	}/* end inner if */
      } /* end outer if */
    }   	
    /* added 3/28/1999 to do copies */ 
   
/*    sginap(1); KLC */
    fflush(NULL);

    /* Make select() do a 1 ms wait - KLC */
    tp.tv_sec = 0;
    tp.tv_usec = 1000;

    /* the following block is non-blocking (no pun intended) */
    /* non-blockingness implemented 3/24/1999 */
   if( select(1, &fds, NULL, NULL, &tp)>0 )    {
       if(!gets(input)){  /* WARNING - commands longer than sizeof(input) will cause unpredictable results */
		/* perror(NULL); */
		/* printf("\nerror is: %s\n", strerror(errno)); */
		goto goback;
	      }else{
		printf("\nno errors happened\n");
	      }
	      /* chomp(input, MAX);  */


    /* scanf("%s\n",input); */
    /* getline(input); */
    /* fgets(input, MAX, stdin); */

   
 
    /* stateless operations */

    if(strncasecmp(input, "help", 4)==0){
	printf(HELP_TEXT);
    }else if(strncasecmp(input,"quit",4)==0){
	done=1;
	 smvrwrite(DEFAULT_FILE, objList);
     	 printf("\nscene written to %s\n", DEFAULT_FILE);

    }else if(strncasecmp(input, "print", 5)==0){ /* print text */
	if( (strlen( input+6 ) + 1) > MAX ) {
	    /* Oops, the text to print is too long for the buffer! */
	    strncpy( Gss->Gtext_buffer, input+6, MAX );
	    Gss->Gtext_buffer[MAX-1] = '\0'; /* Guarantee null terminator on overflow */
	    printf("\n-ERR buffer overflow, first %i characters printed, printed %s\n",  
		   MAX-1,
		   Gss->Gtext_buffer );
	} else {
	    /* Ok, the text will fit */
	    strcpy(Gss->Gtext_buffer, input+6);
	    printf("\n+OK printed %s\n", Gss->Gtext_buffer);
	}
    } else if(ok && strncasecmp(input,"pause",4)==0){
	Gss->Gispaused=1; /* can be unset (released) with the right wand button */ 
	printf("\n+OK pause flag set");
    } else if(ok && strncasecmp(input,"unpause",4)==0){
	Gss->Gispaused=0;         /* can be unset with the right wand button */
	printf("\n+OK pause flag unset");
    } else if(ok && strncasecmp(input,"ispaused",4)==0){
	if(Gss->Gispaused>0){
	    printf("\n+OK ispaused yes");
	}else{
	    printf("\n+OK ispaused no");
	}
    } else switch(s){ 			/*switch on state */
	case LIMBO:
	 if(strncasecmp(input, "init",4)==0){
	   s=INIT;
	  printf("\n+OK system initialized, type help for summary");
	 }else{
	  printf("\n-ERR issue init command\n");
	 }
	break;

	case INIT:
	 ltsize=tokenize(&ltokens, input);
	 
	 /* load the object */
	 if(strncasecmp(ltokens[0], "load", 4)==0){

	  obj = (struct _object *) CAVEMalloc(sizeof(struct _object)); 
	  if (!obj)
                {
                fprintf(stderr,"OUT OF MEMORY!\n");
                return 1;
         }
	 obj->x=0;
	 obj->y=0;
	 obj->z=0;
	 obj->r=0.8;
	 obj->g=0.8;
	 obj->b=1;
	 obj->sx=1;
	 obj->sy=1;
	 obj->sz=1;
	 obj->ro=0;
	 obj->theta=0;
	 obj->phi=0;
	 obj->radius=1;
	 obj->grabbed=0;
	  obj->fChanged = 0;	/* KLC */

   /* physics info added 6/28/1999 by OY */
         obj->mass=1;
         obj->volume=1;
         obj->vx=0;
         obj->vy=0;
         obj->vz=0;
         obj->dt=0.001;




	 strcpy(obj->filename, ltokens[2]);
	 strcpy(obj->name,ltokens[1]);

	 /* now add object to list */
	 obj->next = *objList;
	 *objList = obj;

	 printf("\n+OK loading file %s into %s",ltokens[2],ltokens[1]); 


	}else if(strncasecmp(ltokens[0], "endinit", 4)==0){
	  s=PLAY;
	printf("\n+OK initialization complete");
	  /* now go through all the objects and actually render 
	     them */
	  Gcurrent_list=dlbase;
	 
	  /* iterate through list, render objects */
	/*	 
	 glEnable(GL_LIGHTING);
	  glEnable(GL_LIGHT0); 
	*/
	 /* material commented out 3/25/1999 */
	   glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, whiteMaterial); 
	  for (obj= *objList; obj; obj= obj->next){
	   Gcurrent_list++;
	/* we don't need to load models now, only need to display 
	model=glmReadOBJ(obj->filename);
	   glmUnitize(model);
	   glNewList(Gcurrent_list, GL_COMPILE);
	    glmDraw(model, GLM_TEXTURE|GLM_MATERIAL|GLM_SMOOTH);
	   glEndList();
	   printf("\nloaded file %s into object %s Gcurrent_list is: %d",obj->filename, obj->name,Gcurrent_list);

	*/
	  obj->list_id=Gcurrent_list;
	  obj->display=0;
	  obj->rendered=0;
	  
          }/* end for */
	 } 
	 free2(ltokens,ltsize);
	
	break;

	case PLAY:
         ok=getcmd(input, &cmd);
	 if(!ok){
		/* printf("\nnot OK, we should skip all!\n"); */
	}
          dumpcmd(&cmd); 
	 if(ok && strncasecmp(cmd.cmd,"inv",5)==0){
		if(!smvrinv(objList, cmd.object)){
		 printf("\n-ERR object %s not found",cmd.object);
		}

	}else

	if(ok && strncasecmp(cmd.cmd,"invall",5)==0){
		if(!smvrinvall(objList)){
		 printf("\n-ERR no objects");	
		}



	}else
         if(ok && strncasecmp(cmd.cmd,"write",4)==0){
	   int writeok=1;
	  /* printf("\nCALLING WRITE");  */
	   smvrwrite(cmd.object, objList);
	   if(writeok){
	    printf("\n+OK wrote script to file %s", cmd.object);
	   }else
	    printf("\n-ERR write to file %s failed", cmd.object);
	 }else
	 if(ok && strncasecmp(cmd.cmd,"getpos",4)==0){
		float headPos[3];
		float headAng[3];
		CAVEGetPosition(CAVE_HEAD_NAV, headPos);
		CAVEGetOrientation(CAVE_HEAD_NAV, headAng);
		printf("\n+OK getpos %f %f %f %f %f %f",
		  headPos[0],headPos[1],headPos[2],
		  headAng[0],headAng[1],headAng[2]);

	 }else
	 if(ok && strncasecmp(cmd.cmd,"setpos",4)==0){
		float headPos[3]; 
                float headAng[3];
		switch(cmd.num_args){
                 case 6:
		/* printf("\nnum_args: %d", cmd.num_args); */
		  headPos[0]=cmd.args[2];
		  headPos[1]=cmd.args[3];
		  headPos[2]=cmd.args[4];
		  headAng[0]=cmd.args[5];
                  headAng[1]=cmd.args[6];
                  headAng[2]=cmd.args[7];
		 
                  CAVENavLoadIdentity();
		  CAVENavTranslate(headPos[0],headPos[1],headPos[2]);
		  CAVENavRot(headAng[0], 'x');
		  CAVENavRot(headAng[1],'y');
		  CAVENavRot(headAng[2],'z');
		  printf("\n+OK head pos set to %f %f %f %f %f %f",
			 headPos[0],headPos[1], headPos[2],
			 headAng[0],headAng[1],headAng[2]);
		 break;
	         default:
		  printf("\n-ERR setpos needs 6 args, you gave %d",cmd.num_args);
		 break;
 	        }
		
	 }
	
	 if(ok && strncasecmp(cmd.cmd,"physicson",9)==0){
               Gss->Gphysics=1; 
		printf("\n+OK physicson Newton now rules");
         }
	
 	if(ok && strncasecmp(cmd.cmd,"physicsoff",9)==0){
               Gss->Gphysics=0; 
	       Gss->Ggravity=0;
		printf("\n+OK physicsoff Newton sleeps with da fishes");
         }
 	if(ok && strncasecmp(cmd.cmd,"gravityon",9)==0){
	       if(Gss->Gphysics){
                Gss->Ggravity=1; 
		printf("\n+OK gravityon at 32 feet/sec^2"); 
	       }else{
		printf("\n-ERR gravityon requires physics, turn physics on first ");
	       }
         }
 	if(ok && strncasecmp(cmd.cmd,"gravityoff",9)==0){
               Gss->Ggravity=0; 
		printf("\n+OK gravityoff we can fly");
         }
 	
	if(ok && strncasecmp(cmd.cmd,"funkyon",7)==0){
               Gss->Gfunky=1; 
		printf("\n+OK funkyon watch them cows flip");
         }
	
	if(ok && strncasecmp(cmd.cmd,"funkyoff",7)==0){
               Gss->Gfunky=0; 
		printf("\n+OK funkyoff no more funky stuff going on");
         }
	
 
         if(ok && strncasecmp(cmd.cmd,"display",4)==0){
        	switch(cmd.num_args){
		 case 0:
		    obj=smvrfind(objList, cmd.object);
		    if(obj){ 
		     obj->display=1;
		     printf("\n+OK displaying %s at %f %f %f",
			obj->name, obj->x, obj->y, obj->z);
		    }else{
		     printf("\n-ERR object not found");
		    }
		 break; 
  		 case 3: 	 
		   obj=smvrfind(objList, cmd.object);
		   if(obj){
		    obj->x=cmd.args[2];
		    obj->y=cmd.args[3];
		    obj->z=cmd.args[4];
		    obj->display=1;
		    printf("\n+OK displaying %s at %f %f %f",
			   obj->name, obj->x, obj->y, obj->z);
		   }else{
		     printf("\n-ERR object not found");
		   }


		 break;

		 case 6:
		   obj=smvrfind(objList, cmd.object);
                   if(obj){
                    obj->x=cmd.args[2];
                    obj->y=cmd.args[3];
                    obj->z=cmd.args[4];
		    obj->sx=cmd.args[5];
		    obj->sy=cmd.args[6];
                    obj->sz=cmd.args[7];
  		    obj->radius=smvrmax(obj->sx, obj->sy, obj->sz);	
                    obj->display=1;
                    printf("\n+OK displaying %s at %f %f %f scale %f %f %f",
                           obj->name, obj->x, obj->y, obj->z,
				      obj->sx, obj->sy, obj->sz);
                   }else{
                     printf("\n-ERR object not found");
                   }



		 break;

		 case 9:
 		obj=smvrfind(objList, cmd.object);
                   if(obj){
                    obj->x=cmd.args[2];
                    obj->y=cmd.args[3];
                    obj->z=cmd.args[4];
                    obj->sx=cmd.args[5];
                    obj->sy=cmd.args[6];        
                    obj->sz=cmd.args[7];       
		    obj->radius=smvrmax(obj->sx, obj->sy, obj->sz);
		    obj->ro=cmd.args[8];
                    obj->theta=cmd.args[9];        
                    obj->phi=cmd.args[10]; 

                    obj->display=1;
                    printf("\n+OK displaying %s at %f %f %f scale %f %f %f rot %f %f %f",
                           obj->name, obj->x, obj->y, obj->z,
                                      obj->sx, obj->sy, obj->sz,
				      obj->ro, obj->theta, obj->phi);
                   }else{
                     printf("\n-ERR object not found");
                   }

		 break;


		 case 12:
	    	  obj=smvrfind(objList, cmd.object);
                   if(obj){
                    obj->x=cmd.args[2];
                    obj->y=cmd.args[3];
                    obj->z=cmd.args[4];
                    obj->sx=cmd.args[5];
                    obj->sy=cmd.args[6];        
                    obj->sz=cmd.args[7];      
		    obj->radius=smvrmax(obj->sx, obj->sy, obj->sz); 
                    obj->ro=cmd.args[8];
                    obj->theta=cmd.args[9];        
                    obj->phi=cmd.args[10]; 
		    obj->r=cmd.args[11];
		    obj->g=cmd.args[12]; 
		    obj->b=cmd.args[13];
                    obj->display=1;
                    printf("\n+OK displaying %s at %f %f %f scale %f %f %f rot %f %f %f color %f %f %f",
                           obj->name, obj->x, obj->y, obj->z,
                                      obj->sx, obj->sy, obj->sz,
                                      obj->ro, obj->theta, obj->phi,
				      obj->r, obj->g, obj->b);
                   }else{
                     printf("\n-ERR object not found");
                   }

		 break;

		 default:
		  printf("\n-ERR improper num of args for display\n");


		} /* end switch */
         }else
	 if(ok && strncasecmp(cmd.cmd,"undisplay",4)==0){
	        if(cmd.num_args<=0){	
		 printf("\n+OK undisplaying %s\n",cmd.object);
		obj=smvrfind(objList, cmd.object);
		 if(obj){
		  obj->display=0;
		 }else{
		  printf("\n-ERR object not found"); 
		 }

		}else{
		 printf("\n-ERR improper number of args for undisplay");
	  	}	

	 }else
	 if(ok && strncasecmp(cmd.cmd,"locate",4)==0){
		 obj=smvrfind(objList, cmd.object);
 		 switch(cmd.num_args){
                 case 3: 
		  if(obj){
                    obj->x=cmd.args[2];
                    obj->y=cmd.args[3];
                    obj->z=cmd.args[4];
                   }else{
                     printf("\n-ERR object not found");
                   }
		 break;
		 default:
		   printf("\n-ERR incorrect number of arguments for locate");
		break;	
	       } /* end switch */
	 }else
	 if(ok && strncasecmp(cmd.cmd,"color",4)==0){
		 obj=smvrfind(objList, cmd.object);
                 switch(cmd.num_args){
                 case 3:
                  if(obj){
                    obj->r=cmd.args[2];
                    obj->g=cmd.args[3];
                    obj->b=cmd.args[4];
                   }else{
                     printf("\n-ERR object not found");
                   }
                 break;
                 default:
                   printf("\n-ERR incorrect number of arguments for color");
                break;
               } /* end switch */



	 }else
	 if(ok && strncasecmp(cmd.cmd,"rotate",4)==0){
 		obj=smvrfind(objList, cmd.object);
                 switch(cmd.num_args){
                 case 3:
                  if(obj){
                    obj->ro=cmd.args[2];
                    obj->theta=cmd.args[3];
                    obj->phi=cmd.args[4];
                   }else{
                     printf("\n-ERR object not found");
                   }
                 break;
                 default:
                   printf("\n-ERR incorrect number of arguments for rotate");
                break;
               } /* end switch */

	 }else
	 if(ok && strncasecmp(cmd.cmd,"scale",4)==0){
	   obj=smvrfind(objList, cmd.object);
                 switch(cmd.num_args){
                 case 3:
                  if(obj){
                    obj->sx=cmd.args[2];
                    obj->sy=cmd.args[3];
                    obj->sz=cmd.args[4];
		    obj->radius=smvrmax(obj->sx, obj->sy, obj->sz);
                   }else{
                     printf("\n-ERR object not found");
                   }
                 break;
                 default:
                   printf("\n-ERR incorrect number of arguments for scale");
                break;
               } /* end switch */



	 }else
	 if(ok && strncasecmp(cmd.cmd,"move",4)==0){
	 	switch(cmd.num_args){
		 case 3:
		  
		 break;

		 case 6:

		 break;

		} /* end switch */
	 } else
	 if(ok && strncasecmp(cmd.cmd,"pollchanged",5)==0){
	   /* pollchanged command added by KLC */
	   smvrpollchanged( objList );
	 } else
         if(ok && strncasecmp(cmd.cmd, "clone", 4)==0){
		if(cmd.object2!=NULL){ 
		 struct _object* sourceobj; 
		 printf("\n+OK clone creating %s from %s", cmd.object2, cmd.object);
		 sourceobj=smvrfind(objList, cmd.object2);
 	         if(sourceobj!=NULL){
                  printf("\n^ADDING NEW COPY OBJECT");
                  obj = (struct _object *) CAVEMalloc(sizeof(struct _object));
                  if (!obj){
                   fprintf(stderr,"OUT OF MEMORY!\n");
                  }
                  obj->x = sourceobj->x;
         obj->y = sourceobj->y;
         obj->z = sourceobj->z;
         obj->radius = sourceobj->radius;

         obj->sx=sourceobj->sx;
         obj->sy=sourceobj->sy;
         obj->sz=sourceobj->sz;
         obj->ro=sourceobj->ro;
         obj->theta=sourceobj->theta;
         obj->phi=sourceobj->phi;
         obj->r=sourceobj->r;
         obj->g=sourceobj->g;
         obj->b=sourceobj->b;

         strcpy(obj->filename, sourceobj->filename);
	 strcpy(obj->name, cmd.object);		/*OY only part different from check_add */
          obj->list_id=sourceobj->list_id;
          obj->display=1;
          obj->rendered=1;
         printf("\n^added NEW object %s at %f %f %f l",
                obj->name,obj->x,obj->y,obj->z);
        obj->next = *objList; 
         *objList = obj;
	}
       } /* end if for cmd length checking */
       else{
	 printf("\n-ERR clone does not have enough args");
	}	
      }/* end if for clone */ 




 
 	break; 			/* end state PLAY */
    } /* end outer switch */
    printf("\nsmvr>");
   }  /* end non-blocking block */

   } /* end outer while */ 



  return 0;
} /* end main */

/* read_wand added 6/23/1999 by OY to measure wand velocity
   requires globals Gptime, Gpx, Gpy, Gpz
*/

void read_wand(){
 
 struct timeval thistime;
 float dt, x, y, z;
 float wandPos[3];
 
 char buf[MAX];
 thistime.tv_sec=0;
 thistime.tv_usec=0;
 gettimeofday(&thistime); 
 CAVEGetPosition(CAVE_WAND_NAV,wandPos);
 
 /* printf("\nsec: %d usec: %d", thistime.tv_sec, thistime.tv_usec);  */
 Gss->Gdt=(thistime.tv_sec-Gptime.tv_sec)+(thistime.tv_usec-Gptime.tv_usec)/(float)1000000;
 /* printf("\ndt: %f", dt); */
  
 /* we don't want to divide by zero */ 
 if(Gss->Gdt==0){
  /* sprintf(buf, "avoiding possible divide by zero"); 
   printf("%s\n",buf);
   render2(buf); */
  return;
 }
 

 /* x= wandPos[0];
 y= wandPos[1];
 z= wandPos[2]; */

 Gss->Gvx= (wandPos[0]-Gpx)/Gss->Gdt; /* v= dx/dt */
 Gss->Gvy= (wandPos[1]-Gpy)/Gss->Gdt;
 Gss->Gvz= (wandPos[2]-Gpz)/Gss->Gdt;
 

/*
 if(Gvx!=0||Gvy!=0||Gvz!=0){
  sprintf(buf, "vx=%f vy=%f vz=%f dt=%f", Gvx, Gvy, Gvz, dt);
  printf("%s\n",buf); 
  render2(buf);  
 }
*/

 Gptime.tv_sec=thistime.tv_sec;
 Gptime.tv_usec=thistime.tv_usec;
 Gpx=wandPos[0];
 Gpy=wandPos[1];
 Gpz=wandPos[2];;
 /* sginap(100); */
}


void render(int list, char* string, int size){
        int i; 
        glNewList(list, GL_COMPILE_AND_EXECUTE);
        for(i=0;i<size;i++){
                glutBitmapCharacter(GLUT_BITMAP_8_BY_13, (int)string[i]);
        }
        glEndList();
}

void render2(char* string){
	char retchar='|';
        char ch=string[0];
        int i=0;
	float dy=0.05;	/* distance between lines */
	float dx=0.05;	/* kerning (distance between chars */
	float left=-1.0; /* Left edge of text */ /* KLC */
	int column=0;
	int line=0;	


	/* reserved list called MAX for text */
        glNewList(MAX, GL_COMPILE_AND_EXECUTE);
       /*  glPushMatrix();
        glPushAttrib(1);
         glTranslatef(0,2.45,0);  
	 glRasterPos2d(0,0);  */
	/*	left = 0 - (strchr( string, retchar ) - string) * dx / 2;  Figure out how to center text KLC */
        while(ch!=0){	
		if(ch==retchar){
			line++;
			column=0;
			i++;
			/*			left = 0 - (strchr( string+i, retchar ) - string+i) * dx / 2;  Figure out how to center text KLC */
		}else{


		  glPushMatrix();
		  glPushAttrib(1);
		  glTranslatef(left+(dx*column), 2.45-(dy*line),0);
		  glRasterPos2d(0,0);
		 /*printf("\nat position %f %f %f", 0, 2.45-dy*line, 0); */
                  /* glutBitmapCharacter(GLUT_BITMAP_8_BY_13, (int)ch); */
		   glScalef(3,3,3);
		   glutBitmapCharacter(GLUT_BITMAP_9_BY_15,(int)ch);
		  column++;

		  glPopMatrix();
		  glPopAttrib();
                  i++;
		}
                  ch=string[i];
        }
        /* glPopMatrix();
        glPopAttrib(); */
        glEndList();
                

}

int toggle(int input){
 sginap(100);
 if(input>0){
   input=0;
  }else{
   input=1;
  }
 return input;
}

 
void draw_walls(){
 int i,j,k;
 static int wallID;
 static int firstrun=1;
/* Draw holodeck walls - KLC */
if(firstrun){
 wallID=glGenLists(1);
 glNewList(wallID, GL_COMPILE_AND_EXECUTE); 
 firstrun=0;
 
/* glPushMatrix();
 CAVENavTransform(); */
 Material[0]=0.9;
 Material[1]=0.2;
 Material[2]=0.1;
 glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, Material);
 glPolygonMode( GL_FRONT_AND_BACK, GL_LINE );

 glPushMatrix();
 glTranslatef( 0, 0, 10 );
 for( i = -10; i < 10; i+=1 ) {
   glBegin( GL_QUAD_STRIP );
   for( j = 0; j <= 10; j+=1 ) {
     glVertex3f( i, j, 0 );
     glVertex3f( i+1, j, 0 );
   }
   glEnd( );
 }
 glPopMatrix();

 glPushMatrix();
 glTranslatef( 0, 0, -10 );
 glRotatef( 180, 0, 1, 0 );
 for( i = -10; i < 10; i+=1 ) {
   glBegin( GL_QUAD_STRIP );
   for( j = 0; j <= 10; j+=1 ) {
     glVertex3f( i, j, 0 );
     glVertex3f( i+1, j, 0 );
   }
   glEnd( );
 }
 glPopMatrix();


glPushMatrix();
 glTranslatef( 10, 0, 0 );
 glRotatef( 90, 0, 1, 0 );
 for( i = -10; i < 10; i+=1 ) {
   glBegin( GL_QUAD_STRIP );
   for( j = 0; j <= 10; j+=1 ) {
     glVertex3f( i, j, 0 );
     glVertex3f( i+1, j, 0 );
   }
   glEnd( );
 }
 glPopMatrix();

 glPushMatrix();
 glTranslatef( -10, 0, 0 );
 glRotatef( 270, 0, 1, 0 );
 for( i = -10; i < 10; i+=1 ) {
   glBegin( GL_QUAD_STRIP );
   for( j = 0; j <= 10; j+=1 ) {
     glVertex3f( i, j, 0 );
     glVertex3f( i+1, j, 0 );
   }
   glEnd( );
 }
 glPopMatrix();

 glPushMatrix();
/* glTranslatef( 0, , 0 ); */
 glRotatef( 90, 1, 0, 0 );
 for( i = -10; i < 10; i+=1 ) {
   glBegin( GL_QUAD_STRIP );
   for( j = -10; j <= 10; j+=1 ) {
     glVertex3f( i, j, 0 );
     glVertex3f( i+1, j, 0 );
   }
   glEnd( );
 }
 glPopMatrix();

 glPushMatrix();
 glTranslatef( 0, 10, 0 );
 glRotatef( 90, 1, 0, 0 );
 for( i = -10; i < 10; i+=1 ) {
   glBegin( GL_QUAD_STRIP );
   for( j = -10; j <= 10; j+=1 ) {
     glVertex3f( i, j, 0 );
     glVertex3f( i+1, j, 0 );
   }
   glEnd( );
 }

 glPopMatrix();

/* glPopMatrix(); */
 glPolygonMode( GL_FRONT_AND_BACK, GL_FILL );
 glEndList();
 }else{
  glPushMatrix();
  CAVENavTransform();
  glCallList(wallID);
  glPopMatrix();
 }

}


#define SPEED 1.0f  /* Max navigation speed in feet per second */

/* navigate - perform the navigation calculations. */
void navigate(void)
{ 
 if(!CAVEBUTTON1 && !CAVEBUTTON3){
  float jx=CAVE_JOYSTICK_X,jy=CAVE_JOYSTICK_Y,dt,t;
  static float prevtime = 0;
  t = CAVEGetTime();
  dt = t - prevtime;
  prevtime = t;
  /* printf("\ndt is: %f",dt); */
  if (fabs(jy)>0.2)
        {
        float wandFront[3];
        CAVEGetVector(CAVE_WAND_FRONT,wandFront);
        CAVENavTranslate(wandFront[0]*jy*Gss->Gspeed*dt, wandFront[1]*jy*Gss->Gspeed*dt, wandFront[2]*jy*Gss->Gspeed*dt);
        /* printf("\nmoving around!!!"); */
        }
  if (fabs(jx)>0.4){
        CAVENavRot(-jx*Gss->Grot_angle*dt,'y');
	/* printf("\nrotating!"); */
    }
 } 
}



/* returns largest of 3 numbers */ 
float smvrmax(float a, float b, float c){
  
  if(a>b){
    if(a>c){ 
     return a;
    } 
  }else if(b>c){
    return b;
  }/*else{*/
    return c;
/*  } */
}

/* prints out object characteristics  */
int smvrinv(struct _object **objList, char* objname){
	struct _object *obj;
	obj=smvrfind(objList,objname);
	if(obj){
	 printf("\n+OK inv %s %s %f %f %f %f %f %f %f %f %f %f %f %f\n",
	obj->name, obj->filename, obj->x, obj->y, obj->z, obj->sx, obj->sy, obj->sz, obj->ro, obj->theta, obj->phi, obj->r, obj->g, obj->b);
	 obj->fChanged = 0;	/* KLC */
   	 return 1;
	}

 return 0;
}

int smvrinvall(struct _object** objList){
 struct _object* obj;
 int returnval=1;
 if(objList){
	returnval=1;
 }else{
  	returnval=0;
 }
 for (obj= *objList; obj; obj= obj->next){
	smvrinv(objList, obj->name);
 }
 return returnval;
}

/* Added by KLC on 4/22/199
 * Prints out list of objects which have chaned since last getpos
 */
void smvrpollchanged(struct _object** objList){
 struct _object* obj;
 printf( "+OK pollchanged " );
 for (obj= *objList; obj; obj= obj->next){
   if( obj->fChanged )
     printf( "%s", obj->name);
 }
 printf( "\n" );
}

/* smvrwrite: writes a file containing smvr script to generate current scene */
int smvrwrite(char* filename, struct _object **list){

 FILE* file;
 struct _object* obj;

 file = fopen(filename, "w");
 if (!file) {
    /* printf("\n-ERR couldn't write to file %s",filename); */
    return 0; 
  }
 
 fprintf(file,"init\n");
 
 /* get the filenames */ 
 
 for(obj=*list; obj; obj=obj->next){   
   /* printf("load %s %s\n",obj->name, obj->filename);  */
   fprintf(file,"load %s %s\n",obj->name, obj->filename);

 }
 
 fprintf(file, "endinit\n");
 
 /* get the display arguments and display */
 for(obj=*list;obj;obj=obj->next){
  fprintf(file,"display %s %f %f %f %f %f %f %f %f %f %f %f %f\n",obj->name,
   obj->x, obj->y, obj->z, obj->sx, obj->sy, obj->sz,
   obj->ro, obj->theta, obj->phi, obj->r, obj->g, obj->b);
 }

 fclose(file);
 return 1;
}



struct _object* smvrfind(struct _object** objList, char* str){

 struct _object* obj;
 for (obj= *objList; obj; obj= obj->next){
  if(strcmp(obj->name, str)==0){
   /* printf("\n^smvrfind found %s, returning it\n", str); */
   return obj;
  }
 }
 return NULL;


}


int getcmd(char* in, cmdstruct *command){

	char** tokens;
  	int num_tokens;
	int i=0;

	num_tokens=tokenize(&tokens, in);
	if(num_tokens==0){
		/* printf("\nerror: no tokens"); */
		return 0;
	}
	command->num_args=num_tokens-2;	
	strcpy(command->cmd, tokens[0]); 

	/* Only strcpy the token if there is one there, otherwise set to null */
	if( num_tokens > 1 )
	    strcpy(command->object, tokens[1]);
	else
	    command->object[0] = '\0';

	/* Only strcpy the token if there is one there, otherwise set to null */
	if( num_tokens > 2 )
	    strcpy(command->object2, tokens[2]);
	else
	    command->object2[0] = '\0';

	for(i=2;i<MAX_ARGS;i++){	/* clear of extra stuff */
		(command->args)[i]=ERR;
	}

	for(i=2;i<num_tokens;i++){
		(command->args)[i]=(float)strtod(tokens[i],NULL);
	}
	
	/* free the memory we alloc'd */
	free2(tokens, num_tokens);
 	return 1;
}

/* chomp() removes carriage returns */
void chomp(char* input, int input_size){
 int i=0;
 for(i=0;i<input_size;i++){
        if(input[i]=='\n'){ 
         input[i]='\0';
        }
 }   
}



/* dumpcmd
   for debugging
*/
void dumpcmd(cmdstruct* command){
	int i=0;
	
	printf("\ncommand: %s",command->cmd);
	printf("\nobject: %s",command->object);
	printf("\nnum_args: %d args:\n", command->num_args);
	/* commented out 3/24/1999 */
  	
	for(i=2;(command->args)[i]!=ERR && i<MAX_ARGS;i++){
		printf("%f ",(command->args)[i]);
	}
	printf("\n");


}


void init_gl(void){
 /* added by Oguz 3/7/1999 */
 /* for smvr, we can't glmDraw anything here */

 GLMmodel* model;
/* 
 float redMaterial[] = { 1, 0, 0, 1 };
 float whiteMaterial[]={1,1,1,1};
 float blueMaterial[]={0,0,1,1};
 float greenMaterial[]={0,1,0,1};
*/
 int i,j,k,current_list;
  glEnable(GL_LIGHT0);
  glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, whiteMaterial); 
  /* sphereObj = gluNewQuadric();  */
/*
 dlbase=glGenLists(Gargc);
*/

 /* printf("\ndlbase in init_gl is: %d\n",dlbase); */

  current_list=dlbase;

/* commented out 3/24/1999

 for(i=1;i<Gargc;i++){
   Gcurrent_list++;  
   model=glmReadOBJ(Gargv[i]);
   glmUnitize(model);
   glNewList(Gcurrent_list, GL_COMPILE);
   glmDraw(model, GLM_TEXTURE|GLM_MATERIAL|GLM_SMOOTH);
   glEndList();
   printf("\nloaded object %s current list is %d",Gargv[i],current_list);
   current_list++;
*/
}

void draw_objects(struct _object** objList){
 struct _object *obj;
 GLMmodel* model;
 static int firstrun=1;
 float i, j, k;
 glClearColor(0., 0., 0., 0.);
 glClear(GL_DEPTH_BUFFER_BIT|GL_COLOR_BUFFER_BIT);
 glEnable(GL_LIGHTING);
 	/* draw text */
 	render2(Gss->Gtext_buffer);	
        /* draw a 'cursor' by OY 2/6/1999 */
        glPushMatrix();
	 /* CAVENavTransform(); */
        CAVEWandTransform();
        /* gluSphere(sphereObj, 0.5, 6,6); */
        /* glCallList(dlbase+1);  */
        /* commented out 2/16/1999 */
        glTranslatef(0.0,0,0.1);
        glRotatef(180,0,1,0);
        glutSolidCone(0.3, 0.7, 4, 3);
        glPopMatrix();

/* Draw navigation sphere 
	glPushMatrix();
	 CAVENavTransform(); 
         Material[0]=0.1;
         Material[1]=0.9;
         Material[2]=0.3;
         glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, Material);

	 glutWireSphere(5,7,5);
        glPopMatrix(); 
removed by KLC on Apr 13 1999 in favor of holodeck walls */

/* Draw holodeck walls - KLC 
   code moved to a function -OY */
 /* printf("\n***Gdraw walls in diplay: %d",Gss->Gdraw_walls); */
 if(Gss->Gdraw_walls){
    draw_walls();
  }
 

 for (obj= *objList; obj; obj= obj->next)
       /* extra check added 3/24/1999 for smvr */
        if(obj->display){  
	/* taken out for testing if (obj->grabbed) */
	if(1==2){
                glPushMatrix();
               	/* taken out for testing 4/6/1999 */ 
		/* CAVEWandTransform(); */

		 CAVENavTransform();
                 /* gluSphere(sphereObj, obj->radius, 8, 8); */
                glEnable(GL_NORMALIZE);
                 /* glScalef(obj->radius, obj->radius, obj->radius); */
  		 glScalef(obj->sx, obj->sy, obj->sz);
 		glColor3f(obj->r, obj->g, obj->b);
                Material[0]=obj->r;
                Material[1]=obj->g;
                Material[2]=obj->b;
                glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, Material);
                glCallList(obj->list_id);
                glDisable(GL_NORMALIZE);
                glPopMatrix();
                } 
        else
                {
                glPushMatrix();
 		CAVENavTransform();

	        /* physics stuff added 6/28/1999 by OY */
		if(Gss->Gphysics){

		/* UNIX-specific time related stuff */	
		 struct timeval thistime;
		 gettimeofday(&thistime);
		
   		/*
		 obj->dt=(thistime.tv_sec-obj->ptime.tv_sec)+
                   (thistime.tv_usec-obj->ptime.tv_usec)/(float)1000000;
		 printf("\ndt for %s is %f", obj->name, obj->dt);
		 obj->ptime.tv_sec=thistime.tv_sec;
		 obj->ptime.tv_usec=thistime.tv_usec;
		*/
	  
		/* end UNIX specific */
	
		 obj->vx*=Gss->Gdamping;
		 obj->vy*=Gss->Gdamping;
		 obj->vz*=Gss->Gdamping;
		/*	
                 obj->x+=obj->vx*obj->dt;
                 obj->y+=obj->vy*obj->dt;
                 obj->z+=obj->vz*obj->dt;
		*/
		
		/* gravity */
		/* obj->vy+=GRAV*0.02;	*/
		if(Gss->Ggravity){
		 /* obj->vy-=sqrt((obj->vy*obj->vy)+2*GRAV*fabs(obj->y)); */
		  obj->vy-=sqrt(2*GRAV*fabs(obj->y));
		}
	 	if(obj->y<=0){
		  obj->y=0;	/* hit the ground  and stop */
		  obj->vy=0;
		  obj->vx=0;
		  obj->vz=0;
	        }	
		  	
                 obj->x+=obj->vx*Gss->Gdt;
                 obj->y+=obj->vy*Gss->Gdt;
                 obj->z+=obj->vz*Gss->Gdt;
		
		/*  printf("\nGdt is: %f", Gss->Gdt); */


		} 
	
			




                glTranslatef(obj->x, obj->y, obj->z);
		/* glRotatef(obj->ro, obj->theta, obj->phi,0); */
		/* rotation to be optimized later, 3/28/1999 */
 		glPushMatrix(); /* for rotate */
		 glRotatef(obj->theta, 0, 1, 0);  
		  glRotatef(obj->ro, 1, 0, 0); 
		  glRotatef(obj->phi, 0, 0, 1); 
                /* gluSphere(sphereObj, obj->radius, 8, 8); */
                glEnable(GL_NORMALIZE);
                /* glScalef(obj->radius, obj->radius, obj->radius); */
		glScalef(obj->sx, obj->sy, obj->sz);
	        if(Gss->Gfunky==1){ 
		 glScalef(1+Gss->Gvx, 1+Gss->Gvy, 1+Gss->Gvz);
		}
                /* glCallList(obj->list_id); */
		/* glutSolidCube(0.5); */
		if(!(obj->rendered)){
		 glNewList(obj->list_id,GL_COMPILE); 
		  model=model=glmReadOBJ(obj->filename);
           	  glmUnitize(model);
		  glmDraw(model, GLM_TEXTURE|GLM_MATERIAL|GLM_SMOOTH);
		 glEndList();
		 obj->rendered=1;
	        }
		glScalef(obj->sx, obj->sy, obj->sz);
		glColor3f(obj->r, obj->g, obj->b);
		Material[0]=obj->r;
		Material[1]=obj->g;
		Material[2]=obj->b;
		glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, Material);	
		glPushAttrib( GL_ALL_ATTRIB_BITS );
	        if(Gss->Gdraw_bound){	
                 glutWireSphere(obj->radius,5, 5);
		} else if(obj->grabbed){
		 glutWireSphere(obj->radius,5, 5);
		 /* axis lines - KLC */
		 glColor3f( 1, 0, 0 );
		 glBegin( GL_LINES );
		 glVertex3f( 0, 0, 0 );
		 glVertex3f( 1, 0, 0 );
		 glEnd();

		 glColor3f( 0, 1, 0 );
		 glBegin( GL_LINES );
		 glVertex3f( 0, 0, 0 );
		 glVertex3f( 0, 1, 0 );
		 glEnd();

		 glColor3f( 0, 0, 1 );
		 glBegin( GL_LINES );
		 glVertex3f( 0, 0, 0 );
		 glVertex3f( 0, 0, 1 );
		 glEnd();
		}
		glPopAttrib();
		glCallList(obj->list_id);
		glPopMatrix(); /* for rotate */
                glDisable(GL_NORMALIZE);
                glPopMatrix();
		/*	
		printf("\ndisplayed %s from file %s at %f %f %f \
			 with list_id %d",
		     obj->name, obj->filename, obj->x, obj->y, obj->z,
		     obj->list_id);
		*/
                }
	} /* end outer if */
}


/* possibly not too useful for smvr, we'll just add spheres
   to verify that the program is working */
void check_add(struct _object **list){
   if(Gcopy_object==1){
  
        struct _object *obj;
	struct _object *sourceobj;

        float wandPos[3],wandFront[3], wandAngle[3];
        Gcopy_object=-1; 	/* indicate object JUST copied */	
				/* (button may still be pressed and we want to
				   avoid multiple copies */

	for (sourceobj = *list; sourceobj; sourceobj = sourceobj->next){
		     if(sourceobj->tocopy==1){	
			/* printf("\n^found object: %s",sourceobj->name); */
                        /* sourceobj->grabbed = 1; */
			break;
                     }
        } 
	if(sourceobj!=NULL){	
	   printf("\n^ADDING NEW COPY OBJECT");
	   obj = (struct _object *) CAVEMalloc(sizeof(struct _object));
         if (!obj)
                {
                fprintf(stderr,"OUT OF MEMORY!\n");
                return;
                } 
         CAVEGetPosition(CAVE_WAND_NAV,wandPos);
         CAVEGetVector(CAVE_WAND_FRONT_NAV,wandFront);
        /*  
	 obj->x = wandPos[0] + wandFront[0]*2;
         obj->y = wandPos[1] + wandFront[1]*2;
         obj->z = wandPos[2] + wandFront[2]*2;
         obj->radius = sourceobj->radius;
	*/
	/*
	 obj->x = sourceobj->x + wandFront[0]*2;
         obj->y = sourceobj->y + wandFront[1]*2;
         obj->z = sourceobj->z + wandFront[2]*2;
         obj->radius = sourceobj->radius;
	*/
	 obj->x = sourceobj->x;
         obj->y = sourceobj->y;
         obj->z = sourceobj->z;
         obj->radius = sourceobj->radius;



	 /* sourceobj->grabbed=0; 
         obj->grabbed = 1; 
 	 */ 

	 obj->sx=sourceobj->sx;
         obj->sy=sourceobj->sy;
         obj->sz=sourceobj->sz;
         obj->ro=sourceobj->ro;
         obj->theta=sourceobj->theta;
         obj->phi=sourceobj->phi;
	 obj->r=sourceobj->r;
         obj->g=sourceobj->g;
         obj->b=sourceobj->b;
	
	/* physics implemented 6.28/1999 */
	obj->mass=sourceobj->mass;
	obj->volume=sourceobj->volume;
	obj->vx=sourceobj->vx;
	obj->vy=sourceobj->vy;
	obj->vz=sourceobj->vz;
	obj->dt=sourceobj->dt;

	 strcpy(obj->filename, sourceobj->filename);
	 sprintf(obj->name, "%s_%d",sourceobj->name, Gextra++);
	obj->list_id=sourceobj->list_id;	
	obj->display=1;
	obj->rendered=1;
	 printf("\n^added NEW object %s at %f %f %f l",
		obj->name,obj->x,obj->y,obj->z);
	obj->next = *list;
         *list = obj;
	

        }	
	else{
	 printf("\n^no object found!");
	}
    }
}

struct _object* update_held(struct _object *obj){
 
	static float pro, ptheta,pphi;
	static float px, py, pz;
	if(Gchanged){
		Gchanged=0;
		pro=obj->ro;
		ptheta=obj->theta;
		pphi=obj->phi;
		px=obj->x;
		py=obj->y;
		pz=obj->z;
	
 	}
	/* 6/28/1999 added support for inheriting wand velocity */
	if(Gss->Gphysics){
	 obj->vx=Gss->Gvx;
	 obj->vy=Gss->Gvy;
	 obj->vz=Gss->Gvz;
	}


	/* 6/23/1999, added scale gimmick to test velocity--doesn't work */
	/*	
	if(smvrmax(Gvx, Gvy, Gvz)>1.0){
                obj->sx+=Gvx/10;
		obj->sy+=Gvy/10;
		obj->sz+=Gvz/10;
	} 
	*/

	/* modified 6/13/1999 to do navigate with object */
if (!CAVEBUTTON1 && !CAVEBUTTON2 && !CAVEBUTTON3)
        {
	float wM[4][4];
        float wandPos[3];
        float wandAngle[3]; /* x y z, ro theta phi, or elevation azimuth roll */
	int i,j,k;
	CAVENavTransform();
        CAVEGetPosition(CAVE_WAND_NAV,wandPos);
	CAVEGetOrientation(CAVE_WAND_NAV, wandAngle); 
	/* CAVEGetMatrix(wM); */
        obj->x = wandPos[0];
        obj->y = wandPos[1];
        obj->z = wandPos[2];
 	obj->ro=wandAngle[0]-Gpro+pro;
        obj->theta=wandAngle[1]-Gptheta+ptheta;
        obj->phi=wandAngle[2]-Gpphi+pphi;
	obj->fChanged = 1;	/* KLC */
	/*
	printf("\n*dumping debug output\n");
  	for(i=0;i<4;i++){
		 printf("%f %f %f %f",wM[0][i],wM[1][i],wM[2][i],wM[3][i]);
	}
	printf("\nx: %f y: %f z: %f ro: %f theta: %f phi: %f",
	 obj->x, obj->y, obj->z, obj->ro, obj->theta, obj->phi);
	*/
        obj->grabbed = 0;
	obj->tocopy =0;
	Gcopy_object=0;
        return NULL;
        }
/* code added 4/5/1999 to copy */

 if(CAVEBUTTON1){
    float wandPos[3];
    float wandAngle[3]; /* x y z, ro theta phi, or elevation azimuth roll */
        CAVENavTransform();
        CAVEGetPosition(CAVE_WAND_NAV,wandPos);
        CAVEGetOrientation(CAVE_WAND_NAV, wandAngle);
        obj->x = wandPos[0];
        obj->y = wandPos[1];
        obj->z = wandPos[2];
        obj->ro=wandAngle[0]-Gpro+pro;
        obj->theta=wandAngle[1]-Gptheta+ptheta;
        obj->phi=wandAngle[2]-Gpphi+pphi;
    obj->fChanged = 1;		/* KLC */
   if (CAVE_JOYSTICK_Y > .5){
        /* obj->radius *= 1.005; */
	   /* obj->sx*=1.005; 
 	   obj->sy*=1.005;
	   obj->sz*=1.005; */
	   /* changed 6/23/1999 by OY to fix scaling speed problem */
	  /* commented out 8/3/1999 to unfix scaling speed problem 
	   obj->sx*=1.000005; 
           obj->sy*=1.000005;
           obj->sz*=1.000005; 
	  */
	   obj->sx*=1.0005;
           obj->sy*=1.0005;
           obj->sz*=1.0005;

	   obj->radius=smvrmax(obj->sx, obj->sy, obj->sz);
    }
   else if (CAVE_JOYSTICK_Y < -.5){
        /* obj->radius /= 1.005; */
	   obj->sx/=1.000005;
	   obj->sy/=1.000005;
	   obj->sz/=1.000005;
 	   obj->radius=smvrmax(obj->sx, obj->sy, obj->sz);
   }

   return obj;
 }

 /* code added to navigate with object 6/13/1999 */
 if(CAVEBUTTON2){
	/* just like first button but without resize */
    float wandPos[3];
    float wandAngle[3]; /* x y z, ro theta phi, or elevation azimuth roll */
        CAVENavTransform();
        CAVEGetPosition(CAVE_WAND_NAV,wandPos);
        CAVEGetOrientation(CAVE_WAND_NAV, wandAngle);
        obj->x = wandPos[0];
        obj->y = wandPos[1];
        obj->z = wandPos[2];
        obj->ro=wandAngle[0]-Gpro+pro;
        obj->theta=wandAngle[1]-Gptheta+ptheta;
        obj->phi=wandAngle[2]-Gpphi+pphi;
    obj->fChanged = 1;          /* KLC */

   return obj;
 }

 if(CAVEBUTTON3){
	float wandPos[3];
    	float wandAngle[3]; /* x y z, ro theta phi, or elevation azimuth roll */
        CAVENavTransform();
        CAVEGetPosition(CAVE_WAND_NAV,wandPos);
        CAVEGetOrientation(CAVE_WAND_NAV, wandAngle);
        obj->x = wandPos[0];
        obj->y = wandPos[1];
        obj->z = wandPos[2];
        obj->ro=wandAngle[0];
        obj->theta=wandAngle[1];
        obj->phi=wandAngle[2];
	obj->fChanged = 1;	/* KLC */

   if(Gcopy_object==0){ /* not copied */
     obj->tocopy=1;
     Gcopy_object=1;
   }
   return obj;	/* make sure this is outside the if block1!! 4/5/1999 */
 }
   printf("\n-ERR INTERNAL ERROR, update_held() should never get here"); 
   return NULL;	/* we should never get here */
}

struct _object* check_grab(struct _object **list){
 /* changed 4_5_1999 to copy */

 /* changed 6/13/1999 to navigate with object */

        struct _object *obj;
        float wandPos[3];
	float wandAng[3];
        float outPos[3];  /* physical wand position */
	CAVENavTransform();
        CAVEGetPosition(CAVE_WAND_NAV,wandPos);
	CAVEGetOrientation(CAVE_WAND_NAV, wandAng);	 /* changed 5/8/1999 */
	Gpro=wandAng[0];
	Gptheta=wandAng[1];
	Gpphi=wandAng[2];
	Gchanged=1;

        for (obj = *list; obj; obj = obj->next)
                {
                float dist;
#define SQR(x) ((x)*(x))
                dist = SQR(obj->x - wandPos[0]) + SQR(obj->y - wandPos[1]) +
                        SQR(obj->z - wandPos[2]);
                if (dist < SQR(obj->radius))
                        {
                        obj->grabbed = 1;
			/* printf("\n^check_grab returning %s",obj->name); */
                        return obj;
                        }
                }
 return NULL;

}
